import logging

from typing import (
    Dict,
    List,
    Union,
    Optional,
    Tuple,
)

from overrides import overrides
import torch
from torch import Tensor

from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.fields import Field, ArrayField, LabelField, TextField
from allennlp.data.instance import Instance

from allennlp.common.file_utils import json_lines_from_file

from allennlp_models.vision.dataset_readers.vision_reader import VisionReader

logger = logging.getLogger(__name__)


@DatasetReader.register("visual-entailment")
class VisualEntailmentReader(VisionReader):
    """
    The dataset reader for visual entailment.
    """

    @overrides
    def _read(self, file_path: str):
        split_prefix = "https://storage.googleapis.com/allennlp-public-data/snli-ve/"
        splits = {
            "dev": split_prefix + "snli_ve_dev.jsonl.gz",
            "test": split_prefix + "snli_ve_test.jsonl.gz",
            "train": split_prefix + "snli_ve_train.jsonl.gz",
        }
        file_path = splits.get(file_path, file_path)
        lines = json_lines_from_file(file_path)
        info_dicts: List[Dict] = list(self.shard_iterable(lines))  # type: ignore

        if self.produce_featurized_images:
            # It would be much easier to just process one image at a time, but it's faster to process
            # them in batches. So this code gathers up instances until it has enough to fill up a batch
            # that needs processing, and then processes them all.
            filenames = [info_dict["Flickr30K_ID"] + ".jpg" for info_dict in info_dicts]

            try:
                processed_images = self._process_image_paths(
                    [self.images[filename] for filename in filenames]
                )
            except KeyError as e:
                missing_filename = e.args[0]
                raise KeyError(
                    missing_filename,
                    f"We could not find an image with the name {missing_filename}. "
                    "Because of the size of the image datasets, we don't download them automatically. "
                    "Please download the images from"
                    "https://storage.googleapis.com/allennlp-public-data/snli-ve/flickr30k_images.tar.gz, "
                    "extract them into a directory, and set the image_dir parameter to point to that "
                    "directory. This dataset reader does not care about the exact directory structure. It "
                    "finds the images wherever they are.",
                )
        else:
            processed_images = [None for _ in range(len(info_dicts))]  # type: ignore

        for info_dict, processed_image in zip(info_dicts, processed_images):
            hypothesis = info_dict["sentence2"]
            answer = info_dict["gold_label"]

            instance = self.text_to_instance(processed_image, hypothesis, answer)
            yield instance

    @overrides
    def text_to_instance(
        self,  # type: ignore
        image: Union[str, Tuple[Tensor, Tensor]],
        hypothesis: str,
        label: Optional[str] = None,
        *,
        use_cache: bool = True,
    ) -> Instance:

        tokenized_hypothesis = self._tokenizer.tokenize(hypothesis)
        hypothesis_field = TextField(tokenized_hypothesis, None)

        fields: Dict[str, Field] = {"hypothesis": hypothesis_field}

        if image is not None:
            if isinstance(image, str):
                features, coords = next(self._process_image_paths([image], use_cache=use_cache))
            else:
                features, coords = image

            fields["box_features"] = ArrayField(features)
            fields["box_coordinates"] = ArrayField(coords)
            fields["box_mask"] = ArrayField(
                features.new_ones((features.shape[0],), dtype=torch.bool),
                padding_value=False,
                dtype=torch.bool,
            )

        if label:
            fields["labels"] = LabelField(label)

        return Instance(fields)

    @overrides
    def apply_token_indexers(self, instance: Instance) -> None:
        instance["hypothesis"].token_indexers = self._token_indexers  # type: ignore



from allennlp_models.rc.modules.seq2seq_encoders.multi_head_self_attention import (
    MultiHeadSelfAttention,
)
from allennlp_models.rc.modules.seq2seq_encoders.qanet_encoder import QaNetEncoder
from allennlp_models.rc.modules.seq2seq_encoders.stacked_self_attention import (
    StackedSelfAttentionEncoder,
)



import csv
from typing import Dict, Optional
import logging
import copy

from overrides import overrides

from allennlp.common.checks import ConfigurationError
from allennlp.common.file_utils import cached_path
from allennlp.common.util import START_SYMBOL, END_SYMBOL
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.fields import TextField
from allennlp.data.instance import Instance
from allennlp.data.tokenizers import Tokenizer, SpacyTokenizer, Token
from allennlp.data.token_indexers import TokenIndexer, SingleIdTokenIndexer

logger = logging.getLogger(__name__)


@DatasetReader.register("seq2seq")
class Seq2SeqDatasetReader(DatasetReader):
    """
    Read a tsv file containing paired sequences, and create a dataset suitable for a
    `ComposedSeq2Seq` model, or any model with a matching API.

    Expected format for each input line: <source_sequence_string>\t<target_sequence_string>

    The output of `read` is a list of `Instance` s with the fields:
        source_tokens : `TextField` and
        target_tokens : `TextField`

    `START_SYMBOL` and `END_SYMBOL` tokens are added to the source and target sequences.

    # Parameters

    source_tokenizer : `Tokenizer`, optional
        Tokenizer to use to split the input sequences into words or other kinds of tokens. Defaults
        to `SpacyTokenizer()`.
    target_tokenizer : `Tokenizer`, optional
        Tokenizer to use to split the output sequences (during training) into words or other kinds
        of tokens. Defaults to `source_tokenizer`.
    source_token_indexers : `Dict[str, TokenIndexer]`, optional
        Indexers used to define input (source side) token representations. Defaults to
        `{"tokens": SingleIdTokenIndexer()}`.
    target_token_indexers : `Dict[str, TokenIndexer]`, optional
        Indexers used to define output (target side) token representations. Defaults to
        `source_token_indexers`.
    source_add_start_token : `bool`, (optional, default=`True`)
        Whether or not to add `start_symbol` to the beginning of the source sequence.
    source_add_end_token : `bool`, (optional, default=`True`)
        Whether or not to add `end_symbol` to the end of the source sequence.
    target_add_start_token : `bool`, (optional, default=`True`)
        Whether or not to add `start_symbol` to the beginning of the target sequence.
    target_add_end_token : `bool`, (optional, default=`True`)
        Whether or not to add `end_symbol` to the end of the target sequence.
    start_symbol : `str`, (optional, default=`START_SYMBOL`)
        The special token to add to the end of the source sequence or the target sequence if
        `source_add_start_token` or `target_add_start_token` respectively.
    end_symbol : `str`, (optional, default=`END_SYMBOL`)
        The special token to add to the end of the source sequence or the target sequence if
        `source_add_end_token` or `target_add_end_token` respectively.
    delimiter : `str`, (optional, default=`"\t"`)
        Set delimiter for tsv/csv file.
    quoting : `int`, (optional, default=`csv.QUOTE_MINIMAL`)
        Quoting to use for csv reader.
    """

    def __init__(
        self,
        source_tokenizer: Tokenizer = None,
        target_tokenizer: Tokenizer = None,
        source_token_indexers: Dict[str, TokenIndexer] = None,
        target_token_indexers: Dict[str, TokenIndexer] = None,
        source_add_start_token: bool = True,
        source_add_end_token: bool = True,
        target_add_start_token: bool = True,
        target_add_end_token: bool = True,
        start_symbol: str = START_SYMBOL,
        end_symbol: str = END_SYMBOL,
        delimiter: str = "\t",
        source_max_tokens: Optional[int] = None,
        target_max_tokens: Optional[int] = None,
        quoting: int = csv.QUOTE_MINIMAL,
        **kwargs,
    ) -> None:
        super().__init__(
            manual_distributed_sharding=True, manual_multiprocess_sharding=True, **kwargs
        )
        self._source_tokenizer = source_tokenizer or SpacyTokenizer()
        self._target_tokenizer = target_tokenizer or self._source_tokenizer
        self._source_token_indexers = source_token_indexers or {"tokens": SingleIdTokenIndexer()}
        self._target_token_indexers = target_token_indexers or self._source_token_indexers

        self._source_add_start_token = source_add_start_token
        self._source_add_end_token = source_add_end_token
        self._target_add_start_token = target_add_start_token
        self._target_add_end_token = target_add_end_token
        self._start_token: Optional[Token] = None
        self._end_token: Optional[Token] = None
        if (
            source_add_start_token
            or source_add_end_token
            or target_add_start_token
            or target_add_end_token
        ):
            # Check that the tokenizer correctly appends the start and end tokens to
            # the sequence without splitting them.
            tokens = self._source_tokenizer.tokenize(start_symbol + " " + end_symbol)
            err_msg = (
                f"Bad start or end symbol ('{start_symbol}', '{end_symbol}') "
                f"for tokenizer {self._source_tokenizer}"
            )
            try:
                start_token, end_token = tokens[0], tokens[-1]
            except IndexError:
                raise ValueError(err_msg)
            if start_token.text != start_symbol or end_token.text != end_symbol:
                raise ValueError(err_msg)

            self._start_token = start_token
            self._end_token = end_token

        self._delimiter = delimiter
        self._source_max_tokens = source_max_tokens
        self._target_max_tokens = target_max_tokens
        self._source_max_exceeded = 0
        self._target_max_exceeded = 0
        self.quoting = quoting

    @overrides
    def _read(self, file_path: str):
        # Reset exceeded counts
        self._source_max_exceeded = 0
        self._target_max_exceeded = 0
        with open(cached_path(file_path), "r") as data_file:
            logger.info("Reading instances from lines in file at: %s", file_path)
            for line_num, row in self.shard_iterable(
                enumerate(csv.reader(data_file, delimiter=self._delimiter, quoting=self.quoting))
            ):
                if len(row) != 2:
                    raise ConfigurationError(
                        "Invalid line format: %s (line number %d)" % (row, line_num + 1)
                    )
                source_sequence, target_sequence = row
                if len(source_sequence) == 0 or len(target_sequence) == 0:
                    continue
                yield self.text_to_instance(source_sequence, target_sequence)
        if self._source_max_tokens and self._source_max_exceeded:
            logger.info(
                "In %d instances, the source token length exceeded the max limit (%d) and were truncated.",
                self._source_max_exceeded,
                self._source_max_tokens,
            )
        if self._target_max_tokens and self._target_max_exceeded:
            logger.info(
                "In %d instances, the target token length exceeded the max limit (%d) and were truncated.",
                self._target_max_exceeded,
                self._target_max_tokens,
            )

    @overrides
    def text_to_instance(
        self, source_string: str, target_string: str = None
    ) -> Instance:  # type: ignore
        tokenized_source = self._source_tokenizer.tokenize(source_string)
        if self._source_max_tokens and len(tokenized_source) > self._source_max_tokens:
            self._source_max_exceeded += 1
            tokenized_source = tokenized_source[: self._source_max_tokens]
        if self._source_add_start_token:
            tokenized_source.insert(0, copy.deepcopy(self._start_token))
        if self._source_add_end_token:
            tokenized_source.append(copy.deepcopy(self._end_token))
        source_field = TextField(tokenized_source)
        if target_string is not None:
            tokenized_target = self._target_tokenizer.tokenize(target_string)
            if self._target_max_tokens and len(tokenized_target) > self._target_max_tokens:
                self._target_max_exceeded += 1
                tokenized_target = tokenized_target[: self._target_max_tokens]
            if self._target_add_start_token:
                tokenized_target.insert(0, copy.deepcopy(self._start_token))
            if self._target_add_end_token:
                tokenized_target.append(copy.deepcopy(self._end_token))
            target_field = TextField(tokenized_target)
            return Instance({"source_tokens": source_field, "target_tokens": target_field})
        else:
            return Instance({"source_tokens": source_field})

    @overrides
    def apply_token_indexers(self, instance: Instance) -> None:
        instance.fields["source_tokens"]._token_indexers = self._source_token_indexers  # type: ignore
        if "target_tokens" in instance.fields:
            instance.fields["target_tokens"]._token_indexers = self._target_token_indexers  # type: ignore



# flake8: noqa: F403
from allennlp_models.structured_prediction.predictors import *
from allennlp_models.structured_prediction.dataset_readers import *
from allennlp_models.structured_prediction.metrics import *
from allennlp_models.structured_prediction.models import *



import logging
from typing import Any, Dict

from overrides import overrides

from allennlp.data.fields import MetadataField, TextField, LabelField
from allennlp.common.file_utils import cached_path
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.instance import Instance
from allennlp.data.token_indexers import PretrainedTransformerIndexer
from allennlp.data.tokenizers import PretrainedTransformerTokenizer

logger = logging.getLogger(__name__)


@DatasetReader.register("transformer_superglue_rte")
class TransformerSuperGlueRteReader(DatasetReader):
    """
    Dataset reader for the SuperGLUE Recognizing Textual Entailment task, to be used with a transformer
    model such as RoBERTa. The dataset is in the JSON Lines format.

    It will generate `Instances` with the following fields:

     * `tokens`, a `TextField` that contains the concatenation of premise and hypothesis,
     * `label`, a `LabelField` containing the label, if one exists.
     * `metadata`, a `MetadataField` that stores the instance's index in the file, the original premise,
       the original hypothesis, both of these in tokenized form, and the gold label, accessible as
       `metadata['index']`, `metadata['premise']`, `metadata['hypothesis']`, `metadata['tokens']`,
       and `metadata['label']`.

    # Parameters

    type : `str`, optional (default=`'roberta-base'`)
        This reader chooses tokenizer according to this setting.
    """

    def __init__(
        self,
        transformer_model_name: str = "roberta-base",
        tokenizer_kwargs: Dict[str, Any] = None,
        **kwargs
    ) -> None:
        super().__init__(
            manual_distributed_sharding=True, manual_multiprocess_sharding=True, **kwargs
        )
        self._tokenizer = PretrainedTransformerTokenizer(
            transformer_model_name,
            add_special_tokens=False,
            tokenizer_kwargs=tokenizer_kwargs,
        )
        self._token_indexers = {
            "tokens": PretrainedTransformerIndexer(
                transformer_model_name, tokenizer_kwargs=tokenizer_kwargs, max_length=512
            )
        }

    @overrides
    def _read(self, file_path: str):
        # if `file_path` is a URL, redirect to the cache
        file_path = cached_path(file_path, extract_archive=True)

        logger.info("Reading file at %s", file_path)
        yielded_relation_count = 0
        from allennlp.common.file_utils import json_lines_from_file

        for relation in self.shard_iterable(json_lines_from_file(file_path)):
            premise = relation["premise"]
            hypothesis = relation["hypothesis"]
            if "label" in relation:
                label = relation["label"]
            else:
                label = None
            index = relation["idx"]

            # todo: see if we even need this to be in a separate method
            instance = self.text_to_instance(index, label, premise, hypothesis)

            yield instance
            yielded_relation_count += 1

    @overrides
    def text_to_instance(
        self,
        index: int,
        label: str,
        premise: str,
        hypothesis: str,
    ) -> Instance:
        tokenized_premise = self._tokenizer.tokenize(premise)
        tokenized_hypothesis = self._tokenizer.tokenize(hypothesis)

        fields = {}

        premise_and_hypothesis = TextField(
            self._tokenizer.add_special_tokens(tokenized_premise, tokenized_hypothesis),
        )
        fields["tokens"] = TextField(premise_and_hypothesis)

        # make the metadata
        metadata = {
            "premise": premise,
            "premise_tokens": tokenized_premise,
            "hypothesis": hypothesis,
            "hypothesis_tokens": tokenized_hypothesis,
            "index": index,
        }
        if label:
            fields["label"] = LabelField(label)
            metadata["label"] = label

        fields["metadata"] = MetadataField(metadata)

        return Instance(fields)

    @overrides
    def apply_token_indexers(self, instance: Instance) -> None:
        instance["tokens"].token_indexers = self._token_indexers



from typing import List, Dict

from overrides import overrides
import numpy

from allennlp.common.file_utils import cached_path
from allennlp.common.util import JsonDict
from allennlp.data import Instance
from allennlp.data.fields import LabelField
from allennlp.predictors.predictor import Predictor


@Predictor.register("vilbert_ir")
class VilbertImageRetrievalPredictor(Predictor):
    def predict(self, image: str, caption: str) -> JsonDict:
        image = cached_path(image)
        return self.predict_json({"caption": caption, "image": image})

    @overrides
    def _json_to_instance(self, json_dict: JsonDict) -> Instance:
        from allennlp_models.vision.dataset_readers.flickr30k import Flickr30kReader

        caption = json_dict["caption"]
        image = cached_path(json_dict["image"])
        if isinstance(self._dataset_reader, Flickr30kReader):
            return self._dataset_reader.text_to_instance(caption, image, use_cache=False)
        else:
            raise ValueError(
                f"Dataset reader is of type f{self._dataset_reader.__class__.__name__}. "
                f"Expected {Flickr30kReader.__name__}."
            )

    @overrides
    def predictions_to_labeled_instances(
        self, instance: Instance, outputs: Dict[str, numpy.ndarray]
    ) -> List[Instance]:
        new_instance = instance.duplicate()
        label = numpy.argmax(outputs["probs"])
        new_instance.add_field("label", LabelField(int(label), skip_indexing=True))
        return [new_instance]



import logging
from typing import List, Dict, Any, Optional

from overrides import overrides
import numpy

from allennlp.models import Model
from allennlp.common.util import JsonDict, sanitize
from allennlp.data import Instance, DatasetReader
from allennlp.predictors.predictor import Predictor

from allennlp.data.fields import SpanField

logger = logging.getLogger(__name__)


@Predictor.register("transformer_qa")
class TransformerQAPredictor(Predictor):
    """
    Predictor for the [`TransformerQA`](/models/rc/models/transformer_qa#transformer_qa) model,
    and any other model that takes a question and passage as input.
    """

    def __init__(self, model: Model, dataset_reader: DatasetReader) -> None:
        super(TransformerQAPredictor, self).__init__(model, dataset_reader)
        self._next_qid = 1

    def predict(self, question: str, passage: str) -> JsonDict:
        """
        Make a machine comprehension prediction on the supplied input.
        See [https://rajpurkar.github.io/SQuAD-explorer/](https://rajpurkar.github.io/SQuAD-explorer/)
        for more information about the machine comprehension task.

        # Parameters

        question : `str`
            A question about the content in the supplied paragraph.

        passage : `str`
            A paragraph of information relevant to the question.

        # Returns

        `JsonDict`
            A dictionary that represents the prediction made by the system.
            The answer string will be under the `"best_span_str"` key.

        """
        return self.predict_json({"context": passage, "question": question})

    def predict_json(self, inputs: JsonDict) -> JsonDict:
        results = self.predict_batch_json([inputs])
        assert len(results) == 1
        return results[0]

    @overrides
    def predictions_to_labeled_instances(
        self, instance: Instance, outputs: Dict[str, numpy.ndarray]
    ) -> List[Instance]:
        new_instance = instance.duplicate()
        span_start = int(outputs["best_span"][0])
        span_end = int(outputs["best_span"][1])

        start_of_context = (
            len(self._dataset_reader._tokenizer.sequence_pair_start_tokens)
            + len(instance["metadata"]["question_tokens"])
            + len(self._dataset_reader._tokenizer.sequence_pair_mid_tokens)
        )

        answer_span = SpanField(
            start_of_context + span_start,
            start_of_context + span_end,
            instance["question_with_context"],
        )
        new_instance.add_field("answer_span", answer_span)
        return [new_instance]

    @overrides
    def _json_to_instance(self, json_dict: JsonDict) -> Instance:
        logger.warning(
            "This method is implemented only for use in interpret modules."
            "The predictor maps a question to multiple instances. "
            "Please use _json_to_instances instead for all non-interpret uses. "
        )
        return self._json_to_instances(json_dict, qid=-1)[0]

    def _json_to_instances(self, json_dict: JsonDict, qid: Optional[int] = None) -> List[Instance]:
        # We allow the passage / context to be specified with either key.
        # But we do it this way so that a 'KeyError: context' exception will be raised
        # when neither key is specified, since the 'context' key is the default and
        # the 'passage' key was only added to be compatible with the input for other
        # RC models.
        # if `qid` is `None`, it is updated using self._next_qid
        context = json_dict["passage"] if "passage" in json_dict else json_dict["context"]
        result: List[Instance] = []
        question_id = qid or self._next_qid

        for instance in self._dataset_reader.make_instances(
            qid=str(question_id),
            question=json_dict["question"],
            answers=[],
            context=context,
            first_answer_offset=None,
            is_training=False,
        ):
            self._dataset_reader.apply_token_indexers(instance)
            result.append(instance)
        if qid is None:
            self._next_qid += 1
        return result

    @overrides
    def _batch_json_to_instances(self, json_dicts: List[JsonDict]) -> List[Instance]:
        instances = []
        for json_dict in json_dicts:
            instances.extend(self._json_to_instances(json_dict))
        return instances

    @overrides
    def predict_batch_json(self, inputs: List[JsonDict]) -> List[JsonDict]:
        instances = self._batch_json_to_instances(inputs)
        result = self.predict_batch_instance(instances)
        assert len(result) == len(inputs)
        return result

    @overrides
    def predict_batch_instance(self, instances: List[Instance]) -> List[JsonDict]:
        outputs = self._model.forward_on_instances(instances)

        # group outputs with the same question id
        qid_to_output: Dict[str, Dict[str, Any]] = {}
        for instance, output in zip(instances, outputs):
            qid = instance["metadata"]["id"]
            output["id"] = qid
            output["context_tokens"] = instance["metadata"]["context_tokens"]
            if instance["metadata"]["answers"]:
                output["answers"] = instance["metadata"]["answers"]
            if qid in qid_to_output:
                old_output = qid_to_output[qid]
                if old_output["best_span_scores"] < output["best_span_scores"]:
                    qid_to_output[qid] = output
            else:
                qid_to_output[qid] = output

        return [sanitize(o) for o in qid_to_output.values()]



from overrides import overrides

from allennlp.common.util import JsonDict
from allennlp.data import Instance
from allennlp.predictors.predictor import Predictor


@Predictor.register("seq2seq")
class Seq2SeqPredictor(Predictor):
    """
    Predictor for sequence to sequence models, including

    - [`ComposedSeq2Seq`](../models/composed_seq2seq.md),
    - [`SimpleSeq2Seq`](../models/simple_seq2seq.md),
    - [`CopyNetSeq2Seq`](../models/copynet_seq2seq.md),
    - [`Bart`](../models/bart.md), and
    - [`T5`](../models/t5.md).
    """

    def predict(self, source: str) -> JsonDict:
        return self.predict_json({"source": source})

    @overrides
    def _json_to_instance(self, json_dict: JsonDict) -> Instance:
        """
        Expects JSON that looks like `{"source": "..."}`.
        """
        source = json_dict["source"]
        return self._dataset_reader.text_to_instance(source)

    @classmethod
    def pretrained_t5_for_generation(cls, model_name: str = "t5-base") -> "Seq2SeqPredictor":
        """
        A helper method for creating a predictor for a pretrained T5 generation model.

        # Examples

        ```python
        from allennlp_models.generation.predictors import Seq2SeqPredictor

        ARTICLE_TO_SUMMARIZE = '''
        summarize: The tower is 324 metres (1,063 ft) tall, about the same height as an 81-storey building,
        and the tallest structure in Paris. Its base is square, measuring 125 metres (410 ft) on each side.
        During its construction, the Eiffel Tower surpassed the Washington Monument to become the tallest
        man-made structure in the world, a title it held for 41 years until the Chrysler Building in
        New York City was finished in 1930. It was the first structure to reach a height of 300 metres.
        Due to the addition of a broadcasting aerial at the top of the tower in 1957, it is now taller
        than the Chrysler Building by 5.2 metres (17 ft). Excluding transmitters, the Eiffel Tower is
        the second tallest free-standing structure in France after the Millau Viaduct.
        '''.strip().replace(
            "\n", " "
        )
        predictor = Seq2SeqPredictor.pretrained_t5_for_generation("t5-small")
        predictor.predict(ARTICLE_TO_SUMMARIZE)
        ```
        """
        from allennlp.data import Vocabulary
        from allennlp.data.tokenizers import PretrainedTransformerTokenizer
        from allennlp.data.token_indexers import PretrainedTransformerIndexer
        from allennlp_models.generation.dataset_readers import Seq2SeqDatasetReader
        from allennlp_models.generation.models import T5

        tokenizer, token_indexer = (
            PretrainedTransformerTokenizer(model_name),
            PretrainedTransformerIndexer(model_name),
        )
        reader = Seq2SeqDatasetReader(
            source_tokenizer=tokenizer,
            source_token_indexers={"tokens": token_indexer},
            source_add_start_token=False,
            source_add_end_token=False,
            target_add_start_token=False,
            target_add_end_token=False,
        )
        vocab = Vocabulary.from_pretrained_transformer(model_name)
        model = T5(vocab, model_name)
        return cls(model, reader)
